#include <math.h>
#include <stdlib.h>

#include <mathconst/mathconst.h>
#include <suplib/lists.h>

#include "spline_coef.h"
#include "bsplineval.h"

static BsplineErr
optimized_algo_b(
  double x         /* The point to evaluate the B-spline at */,
  size_t j_index   /* knot[ j_index-1 ] <= x < knot[ j_index ], j */,
  size_t order     /* the order of the B-spline, n */,
  double* c        /* the B-spline coefficients, subscripted c */,
  double* cdiff                                               ,
  double* v        /* work space of dimension : number order + 1 */,
  double* knot     /* the B-spline knots, subscripted x */,
  double* s        /* s( x ) */,
  double* sdiff    /* d s( x ) / dx */
  );

static size_t init_augmented_t_knots( spline_coef* scoef, double* ptr );
static size_t init_augmented_z_knots( spline_coef* scoef, double* ptr );

static double* allocate_knots( size_t num_order, size_t num_knots ) {

  size_t num = 0;
  double* ptr = NULL;

  num = ( num_knots + 2 * (num_order - 1) );

  num *= sizeof( double );

  ptr = (double*) malloc( num );
  if ( NULL == ptr )
    return NULL;

  return ptr;

}

static double* allocate_d( size_t num_order, size_t num_knots ) {

  double* ptr = NULL;

  size_t num = num_order + num_knots;

  num *= sizeof( double );
  ptr = (double*) malloc( num );

  if ( NULL == ptr )
    return NULL;

  return ptr;

}

/*
 * PROCEDURE
 * BsplineAlgo_a  -- The evaluation of s(x) from its normalized B-spline 
 * representation using Scheme A (repeated convex combinations).
 *
 * DESCRIPTION
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 *
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr
BsplineAlgo_a(
  double x           /* The point to evaluate the B-spline at */,
  size_t j_index  /* knot[ j_index-1 ] <= x < knot[ j_index ], j */,
  size_t order                   /* the order of the B-spline, n */,
  double* c       /* the B-spline coefficients, subscripted c */,
  double* d    /* work space of dim : number of knots + order */,
  double* knot           /* the B-spline knots, subscripted x */,
  double* result                                   /* s(x) */
  ) {
		    
  int i_index;                /* i_index corresponds to the index i in paper */
  int l_index;                /* l_index corresponds to the index l in paper */

  /* 		    
   * Set the initial conditions.
   * Step 2. For i = j, j + 1, ...., j + n - 1 set d  = c
   *                                                i    i
   */		        			    
  for ( i_index = j_index; i_index < j_index + order; i_index++ )
    d[ i_index ] = c[ i_index ];

  /* 
   * The value of s(x) is computed by repeated convex combinations in 
   * Steps 3-5
   */

  /*
   * Step 3. For l = 1, 2, ...., n - 1 execute Step 4
   */
  for ( l_index = 1; l_index < order; l_index++ )

    /* 
     * Step 4. For i = j, j + 1, ..., j + n - 1 - l replace d  by
     *                                                       i
     *
     *        ( x - x	     ) d      + ( x  - x ) d 
     * 	             i - n + l 	i + 1 	   i 	    i
     *        ---------------------------------------
     *                    x  -  x
     *	                   i     i - n + l
     */
    for ( i_index = j_index; i_index < j_index + order - l_index; i_index++ ) {

      int i_minus_order_plus_l = i_index - order + l_index;

      d[ i_index ] = ( x - knot[ i_minus_order_plus_l ] ) * d[ i_index + 1 ] +
	( knot[ i_index ] - x ) * d[ i_index ];

      d[ i_index ] /= ( knot[ i_index ] - knot[ i_minus_order_plus_l ] );

    }

  *result = d[ j_index ];

  return Bspline_Success;

}


/*
 * PROCEDURE
 * BsplineAlgo_b:  The evaluation of s(x) from its normalized B-spline 
 * representation using Scheme B (computation of the basis).
 *
 * DESCRIPTION
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 *
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr
BsplineAlgo_b(
  double x           /* The point to evaluate the B-spline at */,
  size_t j_index  /* knot[ j_index-1 ] <= x < knot[ j_index ], j */,
  size_t order                   /* the order of the B-spline, n */,
  double* c       /* the B-spline coefficients, subscripted c */,
  double* v     /* work space of dimension : number order + 1 */,
  double* knot           /* the B-spline knots, subscripted x */,
  double* result                                   /* s(x) */
  ) {

  int i_index, k_index;
  int j_index_1 = j_index - 1;

  double e[ 4 ];

  /* The case n = 1 is treated seperately */
  if ( 1 == order ) {
    *result = c[ j_index ];
    return Bspline_Success;
  }

  /* Initial conditions for the recurrence are set in Steps 3-5. */

  /* 
   * Step 3. Set e  = x - x   , e  = x  - x  and  v  = 1 / ( x  - x   ).
   *              1        j-1   2    j            1          j    j-1
   */
  e[ 1 ] = x - knot[ j_index_1 ];
  e[ 2 ] = knot[ j_index ] - x;
  v[ 1 ] = 1.0 / ( knot[ j_index ] - knot[ j_index_1 ] );
  
  /* 
   * Step 4. For k = 2, 3, .... n - 1 set v  = e  v    / ( x      - x    ).
   *                                       k    1  k-1      j-1+k    j-1
   */
  for ( k_index = 2; k_index < order; k_index++ )
    v[ k_index ] = e[ 1 ] * v[ k_index - 1 ] / 
      ( knot[ j_index_1 + k_index ] - knot[ j_index_1 ] );

  /*
   * Step 5. Set v  = e  v
   *              n    1  n-1
   */
  v[ order ] = e[ 1 ] * v[ order - 1 ];

  /*
   * B-splines are compute by recurrence in Steps 6-10.
   */

  /*
   * Step 6. For k = 1, 2, ..., n - 2 execute Steps 7 - 9.
   */
  for ( k_index = 1; k_index < order - 1; k_index++ ) {

    /*
     * Step 7. Set e  = x - x          and replace v  by e  v / ( x  - x     ).
     *              3        j - 1 - k              1     2  1     j    j-1-k
     */
    e[ 3 ] = x - knot[ j_index_1 - k_index ];
    v[ 1 ] *= e[ 2 ] / ( knot[ j_index ] - knot[ j_index_1 - k_index ] );

    /* 
     * Step 8. For i = 2, 3, ... n - k - 1 replace v  by
     *                                              i
     *          { e  v    + ( x     - x ) v } / ( x      - x     )
     *             3  i-1      j-1+i       i       j-1+i    j-1-k
     */
    for ( i_index = 2; i_index < order - k_index; i_index++ ) {
      v[ i_index ] = e[ 3 ] * v[ i_index - 1 ] +
	( knot[ j_index_1 + i_index ] - x ) * v[ i_index ];
      v[ i_index ] /= ( knot[ j_index_1 + i_index ] - 
			knot[ j_index_1 - k_index ] );
    }

    /*
     * Step 9. Replace v    by e  v      + ( x        - x ) v
     *                  n-k     3  n-k-1      j-1+n-k        n-k
     */
    v[ order - k_index ] = e[ 3 ] * v[ order - k_index - 1 ] +
      ( knot[ j_index_1 + order - k_index ] - x ) * v[ order - k_index ];

  }

  /*
   * Step 10. Replace v  by e  v
   *                   1     2  1
   */
  v[ 1 ] *= e[ 2 ];


  /*
   *  Compute s( x ).
   *                      j + n - 1
   *                        -----
   *                         \
   *             s( x ) =     )     c  v
   *                         /       i  i - j + 1
   *                        -----
   *                        i = j
   */

  *result = 0.0;
  for ( i_index = j_index; i_index < j_index + order; i_index++ )
    *result += c[ i_index ] * v[ i_index - j_index + 1 ];

  return Bspline_Success;

}


/*
 * PROCEDURE
 * BsplineAlgo_c:  The evaluation of s(t,z) from its normalized B-spline 
 * representation. This function allows the user to select the underlying
 * algorithm used, 'a' or 'b'.
 *
 * DESCRIPTION
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 * 
 * RETURNS
 * The enum BsplineErr.
 *
 * SEE ALSO
 *  Bspline_eval
 */
BsplineErr
BsplineAlgo_c( double theta /* The value of theta to evaluate the B-spline */,
		double z         /* The value of z to evaluate the B-spline */,
		BsplineInput* bspline_input /*All neccessary B-spline input */,
		double* result                          /* s( theta, z ) */ ) {

  size_t order;
  size_t j_index, z_index, theta_index;

  spline_coef* scoef = bspline_input->scoef;
  BsplineEvalSpline* eval_spline = bspline_input->eval_spline;

  /*
   * Step 1.  Determine k and l such that z    <= z <= z  and t    <= t < y
   *                                       k-1          k      l-1         l
   */
  if ( Bspline_Success != 
       Bspline_determine_index( &z, scoef->z_knots, scoef->num_z_knots,
				scoef->zmin, scoef->zmax, &z_index ) )
    return Bspline_OutOfRangeZ;

  if ( Bspline_Success != 
       Bspline_determine_index( &theta, scoef->t_knots, scoef->num_theta_knots,
				scoef->tmin, scoef->tmax, &theta_index ) )
    return Bspline_OutOfRangeT;

  /*
   * Step 2.  Evaluate the B-spline basis v  = N   ( x ) (i=k, k+1, ... k+n-1).
   *                                       i    n i
   *
   * Step 3.  for j = l, l + 1, ..., l + order - 1 form
   *
   *                           k + n - 1
   *                             -----
   *                             \
   *                    d   =     )     v  c
   *                     j       /       i   i j
   *                             -----
   *                             i = k
   *
   * The description for scheme B:
   *
   * i.e. the values of v  = N   ( x )   ( i = k, k + 1, ..., k + n - 1 ),
   *                     i    n i
   *
   * follow by the direct evaluation of
   *
   *                             k + n - 1
   *                               -----
   *                               \
   *                     s(x) =     )     v  c
   *                               /       i  i
   *                               -----
   *                               i = k
   *
   */

  {
    BsplineErr (*ptr1)(double,size_t,size_t,double*,double*,double*,double*) = NULL;
    BsplineErr (*ptr2)(double,size_t,size_t,double*,double*,double*,double*) = NULL;


    if ( 'a' == bspline_input->method1 || 'A' == bspline_input->method1 )
      ptr1 = BsplineAlgo_a;
    else if ( 'b' == bspline_input->method1 || 'B' == bspline_input->method1 )
      ptr1 = BsplineAlgo_b;
    else
      return Bspline_WrongMethod1;

    if ( 'a' == bspline_input->method2 || 'A' == bspline_input->method2 )
      ptr2 = BsplineAlgo_a;
    else if ( 'b' == bspline_input->method2 || 'B' == bspline_input->method2 )
      ptr2 = BsplineAlgo_b;
    else
      return Bspline_WrongMethod2;

    order = scoef->z_order;
    for ( j_index = theta_index; j_index < theta_index + order; j_index++ )
      ptr1( z, z_index, order,
	   scoef->c + (j_index % scoef->num_theta_knots) * scoef->qz,
	   eval_spline->dz, eval_spline->z_knots + scoef->z_order - 1,
	   &(eval_spline->d[ j_index ]) );

    ptr2( theta, theta_index, order, eval_spline->d,
	  eval_spline->dt, eval_spline->t_knots + scoef->theta_order - 1,
	  result );

  }

  return Bspline_Success;

}


/*
 * PROCEDURE
 * BsplineAlgo_c_derivs -- The evaluation of s(t,z) and the derivative with
 * respect to t and z from its normalized B-spline representation. 
 * This function allows the user to select the underlying
 * algorithm used, 'a' or 'b'.
 *
 * DESCRIPTION
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 *
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr
BsplineAlgo_c_derivs( double theta   /* The theta to evaluate the B-spline */,
		       double z  /* The value of z to evaluate the B-spline */,
		       BsplineInput* bspline_input    /* All B-spline input */,
		       BsplineResult* bspline_result /* All B-spline output */)
{

  size_t zorder, torder;
  size_t j_index, z_index, theta_index;
  spline_coef* scoef = bspline_input->scoef;
  BsplineEvalSpline* eval_spline = bspline_input->eval_spline;

  if ( Bspline_Success != 
       Bspline_determine_index( &z, scoef->z_knots, scoef->num_z_knots,
				scoef->zmin, scoef->zmax, &z_index ) )
    return Bspline_OutOfRangeZ;

  if ( Bspline_Success != 
       Bspline_determine_index( &theta, scoef->t_knots, scoef->num_theta_knots,
				scoef->tmin, scoef->tmax, &theta_index ) )
    return Bspline_OutOfRangeT;

  {
    BsplineErr (*ptr1)(double,size_t,size_t,double*,double*,double*,double*) = NULL;
    BsplineErr (*ptr2)(double,size_t,size_t,double*,double*,double*,double*) = NULL;

    if ( 'a' == bspline_input->method1 || 'A' == bspline_input->method1 )
      ptr1 = BsplineAlgo_a;
    else if ( 'b' == bspline_input->method1 || 'B' == bspline_input->method1 )
      ptr1 = BsplineAlgo_b;
    else
      return Bspline_WrongMethod1;

    if ( 'a' == bspline_input->method2 || 'A' == bspline_input->method2 )
      ptr2 = BsplineAlgo_a;
    else if ( 'b' == bspline_input->method2 || 'B' == bspline_input->method2 )
      ptr2 = BsplineAlgo_b;
    else
      return Bspline_WrongMethod2;


    /***************************** s( t, z ) *********************************/
    zorder = scoef->z_order;
    for ( j_index = theta_index; j_index < theta_index + zorder; j_index++ )
      ptr1( z, z_index, zorder,
	    scoef->c + (j_index % scoef->num_theta_knots) * scoef->qz,
	    eval_spline->dz, eval_spline->z_knots + scoef->z_order - 1,
	    &(eval_spline->d[ j_index ]) );

    torder = scoef->theta_order;
    ptr2( theta, theta_index, torder, eval_spline->d, eval_spline->dt, 
	  eval_spline->t_knots + scoef->theta_order - 1, 
	  &bspline_result->s );

    /************************** d s( t, z ) / dt *****************************/
    zorder = scoef->z_order;
    for ( j_index = theta_index; j_index < theta_index + zorder; j_index++ )
      ptr1( z, z_index, zorder,
	    scoef->cdift + (j_index % scoef->num_theta_knots) * scoef->qz,
	    eval_spline->dz, eval_spline->z_knots + scoef->z_order - 1,
	    &(eval_spline->d[ j_index ]) );

    /* derivative wrt theta, therefore the order in theta is one less */
    torder = scoef->theta_order - 1;
    ptr2( theta, theta_index, torder, eval_spline->d, eval_spline->dt, 
	  eval_spline->t_knots + scoef->theta_order - 1, 
	  &bspline_result->dsdt );

    /************************** d s( t, z ) / dz *****************************/
    /* derivative wrt z, therefore the order in z is one less */
    zorder = scoef->z_order - 1;
    /* 
     * note that the for loop has an added one since the algorithm must still
     * be evaluated for zorder number of elements.
     */
    for ( j_index = theta_index; j_index < theta_index + zorder + 1; 
	  j_index++ )
      ptr1( z, z_index, zorder,
	    scoef->cdifz + 
	    (j_index % scoef->num_theta_knots) * (scoef->qz-1),
	    eval_spline->dz, 
	    eval_spline->z_knots + scoef->z_order - 1,
	    &(eval_spline->d[ j_index ]) );

    torder = scoef->theta_order;
    ptr2( theta, theta_index, torder, eval_spline->d, eval_spline->dt, 
	  eval_spline->t_knots + scoef->theta_order - 1, 
	  &bspline_result->dsdz );

  }

  return Bspline_Success;

}

/*
 * PROCEDURE
 * Bspline_err_msg -- print the error message to the file pointer.
 *
 * RETURNS
 * void.
 */
void Bspline_err_msg( BsplineErr err, FILE* fp ) {

  switch( err ) {
  case Bspline_OutOfRange:
    fprintf( fp, "Argument is not within the knots\n" );
    break;
  case Bspline_OutOfRangeT:
    fprintf( fp, "Argument for theta is not within the knots\n" );
    break;
  case Bspline_OutOfRangeZ:
    fprintf( fp, "Argument for z is not within the knots\n" );
    break;
  case Bspline_WrongMethod1:
    fprintf( fp, "The value for argument `method1' must be `a' or `A'\n" );
    break;
  case Bspline_WrongMethod2:
    fprintf( fp, "The value for argument `method2' must be `b' or `B'\n" );
    break;
  default:
    fprintf( fp, "Unrecognized error message\n" );
    break;
  }

  return;

}

/*-----------------------------------------------------------------------*/
static int cmp_knots( const void* key, const void* datum ) {

  double diff = (*(double *) key) - (*(double *) datum );

  if ( diff >= 0.0 )
    return 1;
  else if ( diff < 0.0 )
    return -1;
  else 
    return 0;

}

/*
 * PROCEDURE
 * Determine the index such that base[ index - 1 ] <= key < base[ index ].
 * If key == base[ num - 1 ] then index is set to num - 1.
 *
 * DESCRIPTION
 * Do a Binary search to find the location of the key within the knots.  If 
 * this routine is usually called iteratively between the ideal surface and
 * the deformation map then it can improve a bit by making use of the last
 * position found.
 * 
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr 
Bspline_determine_index( double* key             /* The key to search for */,
			   const void* base   
			   /* Points to the element at the base of array */,
			   size_t num    /* Number of elements in the array */,
			   double minval     /* The left most knot boundary */,
			   double maxval    /* The right most knot boundary */,
			   size_t* index  
			   /* base[ index - 1 ] <= key < base[ index ] */ ) {

  const void* low=NULL;
  const void* high=NULL;
  size_t size = sizeof( double );

  /*
   * Check for the right most boundary condition.
   * index = N if  x == x
   *                      N
   if ( ((double*)base)[ num - 2 ] < *key && 
   *key <= ((double*)base)[ num - 1 ] ) {
   *index = num - 1;
   return Bspline_Success;
   }
   */

 bnd_bsearch( key, base, num, size, &low, &high, &cmp_knots );

  if ( low && high ) {
    /*
     *  (high - base) is used instead of (low - base) because 
     *
     *                    x      <=  x  <  x
     *                     i - 1             i
     */
    *index = (double*) high - (double*) base;
    return Bspline_Success;
  }

  /* 
   * The special cases to check.  One can still evaluate the spline if the
   * points lie at [ min, base[ 0 ] ) and [ base[ num - 1 ], max ]
   */

  if ( minval <= *key && *key < ((double*)base)[ 0 ] ) {
    *index = 0;
    return Bspline_Success;
  }

  if ( ((double*)base)[ num - 1 ] <= *key && *key <= maxval ) { 
    *index = num;
    return Bspline_Success;
  }

  return Bspline_OutOfRange;

}

/*
 * PROCEDURE
 * Bspline_eval:  The evaluation of s(t,z) from its normalized B-spline 
 * representation.  This function is similar to the function
 * BsplineAlgo_c with the exception that the underlying algorithms 'b'
 * and 'b' are chosen for the user.
 *
 * DESCRIPTION
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 * 
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr
Bspline_eval( double theta /* The value of theta to evaluate the B-spline */,
		double z         /* The value of z to evaluate the B-spline */,
		BsplineInput* bspline_input /*All neccessary B-spline input */,
		double* result                          /* s( theta, z ) */ ) {

  bspline_input->method1 = 'b';
  bspline_input->method2 = 'b';

  return BsplineAlgo_c( theta, z, bspline_input, result );

}


/*
 * PROCEDURE
 * Bspline_eval_derivs -- The evaluation of s(t,z) and the derivative
 * with respect to t and z from its normalized B-spline representation.
 * This function is an optimized version of the function 
 * BsplineAlgo_c_derivs.
 * 
 *
 * DESCRIPTION
 * An optimized implementation of the function BsplineAlgo_c.
 * This function can be further optimized by folding in the last of the 
 * seperate call to BsplineAlgo_b, but this requires the rewriting of
 * the algo_b routine.
 * The Numerical Evaluation of a Spline from its B-Spline Representation
 * by M. G. Cox.  J. Inst. Maths. Applics (1978) 21, 135-143.
 *
 * RETURNS
 * The enum BsplineErr.
 */
BsplineErr
Bspline_eval_derivs( double theta     /* The theta to evaluate the B-spline */,
		     double z    /* The value of z to evaluate the B-spline */,
		     BsplineInput* bspline_input      /* All B-spline input */,
		     BsplineResult* bspline_result /* All B-spline output */) {

  size_t j_index, z_index, theta_index;
  spline_coef* scoef = bspline_input->scoef;
  BsplineEvalSpline* eval_spline = bspline_input->eval_spline;

  if ( Bspline_Success != 
       Bspline_determine_index( &z, scoef->z_knots, scoef->num_z_knots, 
				scoef->zmin, scoef->zmax, &z_index ) )
    return Bspline_OutOfRangeZ;

  if ( Bspline_Success != 
       Bspline_determine_index( &theta, scoef->t_knots, scoef->num_theta_knots,
				scoef->tmin, scoef->tmax, &theta_index ) )
    return Bspline_OutOfRangeT;

  /* Step 2 and Step 4 */
  {
    size_t zorder = scoef->z_order;
    for ( j_index = theta_index; j_index < theta_index + zorder; j_index++ ) {
      size_t mod_index = j_index % scoef->num_theta_knots;
      optimized_algo_b( z, z_index, zorder, scoef->c + mod_index * scoef->qz,
			scoef->cdift + mod_index * scoef->qz, eval_spline->dz, 
			eval_spline->z_knots + scoef->z_order - 1,
			&(eval_spline->d[ j_index ]), 
			&(eval_spline->dift[ j_index ] ) );
      BsplineAlgo_b( z, z_index, zorder - 1,
		   scoef->cdifz + mod_index * ( scoef->qz - 1 ),
		   eval_spline->dz, 
		   eval_spline->z_knots + scoef->z_order - 1,
		   &(eval_spline->difz[ j_index ]) );
    }

    /* Step 4. */
    {
      size_t torder = scoef->theta_order;
      double *ptr = eval_spline->t_knots + scoef->theta_order - 1;
      optimized_algo_b( theta, theta_index, torder, eval_spline->d, 
			eval_spline->difz, eval_spline->dt, ptr, 
			&bspline_result->s, 
			&bspline_result->dsdz );

      /* derivative wrt theta, therefore the order in theta is one less */
      torder = scoef->theta_order - 1;
      BsplineAlgo_b( theta, theta_index, torder, eval_spline->dift, 
		   eval_spline->dt, ptr,  &bspline_result->dsdt );
      
    }

  }

  return Bspline_Success;

}

/*
 * PROCEDURE
 * Bspline_free -- free memory allocated by Bspline_init.
 *
 * RETURNS
 * void.
 */
void Bspline_free( BsplineInput* input ) {

  BsplineEvalSpline* ptr = input->eval_spline;
 
  if ( ptr ) {

    if ( ptr->t_knots )
      free( ptr->t_knots );

    if ( ptr->z_knots )
      free( ptr->z_knots );

    if ( ptr->d )
      free( ptr->d );

    if ( ptr->difz )
      free( ptr->difz );

    if ( ptr->dift )
      free( ptr->dift );

    if ( ptr->dt )
      free( ptr->dt );

    if ( ptr->dz )
      free( ptr->dz );

    if ( ptr )
      free( ptr );

  }

  free( input );

}

/*
 * PROCEDURE
 * Allocates the memory used for Bspline library.
 *
 * RETURNS
 * BsplineInput* if successful else NULL is returned.
 */
BsplineInput* Bspline_init( spline_coef* scoef ) {

  size_t nknots=0, norder=0;
  BsplineInput* input = NULL;
  BsplineEvalSpline* ptr = NULL;

  input = (BsplineInput*) malloc( sizeof( BsplineInput ) );
  if ( NULL == input )
    return NULL;

  input->scoef = scoef;

  input->eval_spline = (BsplineEvalSpline*) malloc( sizeof( BsplineEvalSpline ) );
  if ( NULL == input->eval_spline ) {
    free( input );
    return NULL;
  }

  ptr = input->eval_spline;

  /* 
   * initialize to NULL.  This will enable us to free memory already allocated
   * if one has to leave this routine abruptly.
   */
  ptr->d = NULL;
  ptr->difz = NULL;
  ptr->dift = NULL;
  ptr->dt = NULL;
  ptr->dz = NULL;
  ptr->t_knots = NULL;
  ptr->z_knots = NULL;

  nknots = scoef->num_theta_knots > scoef->num_z_knots ? 
    scoef->num_theta_knots : scoef->num_z_knots;
  norder = scoef->theta_order > scoef->z_order ? 
    scoef->theta_order : scoef->z_order;
  if ( NULL == ( ptr->d = allocate_d( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  if ( NULL == ( ptr->difz = allocate_d( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  if ( NULL == ( ptr->dift = allocate_d( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }

  /* Allocate memory for the theta arrays */
  nknots = scoef->num_theta_knots;
  norder = scoef->theta_order;
  if ( NULL == ( ptr->dt = allocate_d( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  if ( NULL == ( ptr->t_knots = allocate_knots( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  init_augmented_t_knots( scoef, ptr->t_knots );

  /* Allocate memory for the z arrays */
  nknots = scoef->num_z_knots;
  norder = scoef->z_order;
  if ( NULL == ( ptr->dz = allocate_d( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  if ( NULL == ( ptr->z_knots = allocate_knots( norder, nknots ) ) ) {
    Bspline_free( input );
    return NULL;
  }
  init_augmented_z_knots( scoef, ptr->z_knots );

  return input;

}

/*
 * Page 11 of the DASL Definitions
 *
 * Periodically extended knots:
 * 
 * For given interior knots knot  , knot  , ..., knot  , contained strictly
 *                              1       2            N
 * whithin the interval t    to t   , the periodically extended knots are
 *                       min     max
 * defined for any integer j by
 *
 *                      knot  = knot     - ( t    - t   )
 *                          j       j+N       max    min
 */
static size_t
init_augmented_t_knots( spline_coef* scoef, double* ptr ) {

  size_t i;
  size_t num_pad = scoef->theta_order - 1;
  double period = scoef->tmax - scoef->tmin;

  /* The interior knots */
  for ( i = 0; i < scoef->num_theta_knots; i++ )
    ptr[ num_pad + i ] = scoef->t_knots[ i ];

  /* Periodic boundary condition */
  for ( i = 0; i < num_pad; i++ )
    ptr[ i ] = ptr[ i + scoef->num_theta_knots ] - period;

  /* Periodic boundary condtion */
  for ( i = num_pad + scoef->num_theta_knots;
	i < 2 * num_pad + scoef->num_theta_knots; i++ )
    ptr[ i ] = ptr[ i - scoef->num_theta_knots ] + period;

  /*
  for ( i = 0; i < scoef->num_theta_knots + 2 * num_pad; i++ )
    if ( 0 == i || scoef->num_theta_knots + 2 * num_pad - 1 == i )
      fprintf( stdout, "#t_knot[ %d ] = %g\n", i, ptr[ i ] );
    else
      fprintf( stdout, "#t_knot[ %d ] = %g\tdelta = %g\n", 
	       i, ptr[ i ], ptr[ i ] - ptr[ i - 1 ] );
  exit( 0 );
  */

  return EXIT_SUCCESS;

}

static size_t init_augmented_z_knots( spline_coef* scoef, double* ptr ) {

  int i, index = 0;
  int num_pad = scoef->z_order - 1;

  double first = scoef->zmin;
  double last = scoef->zmax;

  /* Coincident points at the left boundary, set to first knot */
  for ( i = 0; i < num_pad; i++ )
    ptr[ index++ ] = first;

  /* The interior knots */
  for ( i = 0; i < scoef->num_z_knots; i++ )
    ptr[ index++ ] = scoef->z_knots[ i ];
 
  /* The coincident points at the right boundary, set to last knot */
  for ( i = 0; i < num_pad; i++ )
     ptr[ index++ ] = last;

  /*
  for ( i = 0; i < scoef->num_z_knots + 2 * num_pad; i++ )
    fprintf( stdout, "#z_knot[ %d ] = %g\n", i, ptr[ i ] );
    */

  return EXIT_SUCCESS;

}

/* 
 * Same as BsplineAlgo_b above but optimized for performance.
 */
static BsplineErr
optimized_algo_b( double x         /* The point to evaluate the B-spline at */,
		  size_t j_index/* knot[ j_index-1 ] <= x < knot[ j_index ], j */,
		  size_t order                 /* the order of the B-spline, n */,
		  double* c     /* the B-spline coefficients, subscripted c */,
		  double* cdiff                                               ,
		  double* v   /* work space of dimension : number order + 1 */,
		  double* knot         /* the B-spline knots, subscripted x */,
		  double* s                                       /* s( x ) */,
		  double* sdiff                         /* d s( x ) / dx */ ) {

  size_t i_index, k_index;
  size_t j_index_1 = j_index - 1;

  double e[ 4 ];

  if ( 1 == order ) {
    *s = c[ j_index ];
    *sdiff = cdiff[ j_index ];
    return Bspline_Success;
  }

  e[ 1 ] = x - knot[ j_index_1 ];
  e[ 2 ] = knot[ j_index ] - x;
  v[ 1 ] = 1.0 / ( knot[ j_index ] - knot[ j_index_1 ] );
  
  for ( k_index = 2; k_index < order; k_index++ )
    v[ k_index ] = e[ 1 ] * v[ k_index - 1 ] / 
      ( knot[ j_index_1 + k_index ] - knot[ j_index_1 ] );

  v[ order ] = e[ 1 ] * v[ order - 1 ];

  for ( k_index = 1; k_index < order - 1; k_index++ ) {

    size_t order_k_index = order - k_index;

    e[ 3 ] = x - knot[ j_index_1 - k_index ];

    v[ 1 ] *= e[ 2 ] / ( knot[ j_index ] - knot[ j_index_1 - k_index ] );

    for ( i_index = 2; i_index < order_k_index; i_index++ ) {
      v[ i_index ] = e[ 3 ] * v[ i_index - 1 ] +
	( knot[ j_index_1 + i_index ] - x ) * v[ i_index ];

      v[ i_index ] /= ( knot[ j_index_1 + i_index ] - 
			knot[ j_index_1 - k_index ] );

    }

    order_k_index = order - k_index;

    v[ order_k_index ] = e[ 3 ] * v[ order_k_index - 1 ] +
      ( knot[ j_index_1 + order_k_index ] - x ) * v[ order_k_index ];

  }

  v[ 1 ] *= e[ 2 ];

  *s = 0.0;
  *sdiff = 0.0;
  for ( i_index = j_index; i_index < j_index + order; i_index++ ) {
    double v_index = v[ i_index - j_index + 1 ];
    *s += c[ i_index ] * v_index;
    *sdiff += cdiff[ i_index ] * v_index;
  }

  return Bspline_Success;

}

